/*
 *  RapidMiner
 *
 *  Copyright (C) 2001-2015 by RapidMiner GmbH and the contributors
 *
 *  Complete list of developers available at our web site:
 *
 *       http://rapidminer.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses/.
 */
package cl;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.table.AttributeFactory;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.learner.PredictionModel;
import com.rapidminer.tools.Ontology;
import com.rapidminer.tools.Tools;

/**
 * The model (in the sense of Rapidminer) for the string coordinate.
 * 
 * @author Christian Leitold
 */
public class StringCoordinateModel extends PredictionModel {

	private static final long serialVersionUID = 8381268071090932037L;

	private String[] attributeNames;
	
	private String[] attributeConstructions;
		
	private boolean[] selectedAttributes;
		
	private String fixedAttribute;
		
	private StringCoordinateModelParameters modelParameters;
	
	private boolean useCommittorData;
	
	private StringEvaluator stringEvaluator;
	
	public StringCoordinateModel(ExampleSet exampleSet, boolean[] selectedAttributes, String fixedAttribute, StringCoordinateModelParameters modelParameters, boolean useCommittorData) {
		super(exampleSet);
		this.attributeNames = com.rapidminer.example.Tools.getRegularAttributeNames(exampleSet);
		this.attributeConstructions = com.rapidminer.example.Tools.getRegularAttributeConstructions(exampleSet);
		this.selectedAttributes = selectedAttributes;
		this.fixedAttribute = fixedAttribute;
		this.modelParameters = modelParameters;
		this.useCommittorData = useCommittorData;
		this.stringEvaluator = new StringEvaluator(modelParameters);
	}

	@Override
	public ExampleSet performPrediction(ExampleSet exampleSet, Attribute predictedLabel) throws OperatorException {
		
		Attribute[] attributes = new Attribute[attributeNames.length];
		
		int usedAttributeCounter = 0;
		
		for (int i = 0; i < attributeNames.length; i++) {
			attributes[i] = exampleSet.getAttributes().get(attributeNames[i]);
			
			if (selectedAttributes[i]) usedAttributeCounter++; 
			
			if (attributes[i] == null && selectedAttributes[i])
				throw new UserError(null, 111, attributeNames[i]);
		}
		
		int numberOfUsedAttributes = usedAttributeCounter;
		int attributeCounter;
		
		Attribute reactionCoordinate = null;
		if (useCommittorData) {
			reactionCoordinate = AttributeFactory.createAttribute("reaction_coordinate", Ontology.REAL);
			exampleSet.getExampleTable().addAttribute(reactionCoordinate);
			exampleSet.getAttributes().setSpecialAttribute(reactionCoordinate, "reaction_coordinate");
		}
		
		double[] inverseScalingFactors = new double[numberOfUsedAttributes];
		double[] scalingFactors = modelParameters.getScalingFactors();
		for (int i=0; i<numberOfUsedAttributes; i++) {
			inverseScalingFactors[i] = 1.0d / scalingFactors[i];
		}
		
		for (Example example : exampleSet) {
			
			attributeCounter = 0;
			usedAttributeCounter = 0;
			double[] data = new double[numberOfUsedAttributes];
			for (Attribute attribute : attributes) {
				if (selectedAttributes[attributeCounter]) {
					data[usedAttributeCounter] = example.getValue(attribute) * scalingFactors[usedAttributeCounter];
					usedAttributeCounter++;
				}
				attributeCounter++;
			}
			
        	double prediction = stringEvaluator.getStringPrediction(data);
        	
        	if (useCommittorData) {
        		double pB = 0.5d + 0.5d*Math.tanh(prediction);
        		example.setValue(reactionCoordinate, prediction);
        		example.setValue(predictedLabel, pB);
        	}
        	else {
        		example.setValue(predictedLabel, prediction);
        	}
        	
		}
		
		return exampleSet;
	}
	
	// print human readable representation of the model
	@Override
	public String toString() {
		
		StringBuffer result = new StringBuffer();
		
		result.append("StringCoordinateModel text representation" + Tools.getLineSeparator());
		result.append("*****************************************" + Tools.getLineSeparator() + Tools.getLineSeparator());
		
		result.append(String.format("String dimension m = %d", modelParameters.getMStringDimension()) + Tools.getLineSeparator());
		result.append(String.format("String length M = %d (+2)", modelParameters.getMStringPoints()) + Tools.getLineSeparator());
		result.append(Tools.getLineSeparator());
		
		int usedCounter = 0;
		for (int i=0; i<selectedAttributes.length; i++) {
			if (selectedAttributes[i]) {
				result.append("used attribute " + Integer.toString(usedCounter) + " = " + attributeConstructions[i] + Tools.getLineSeparator());
				usedCounter++;
			}
		}
		result.append(Tools.getLineSeparator());
		
		result.append(String.format("fixed index = %d", modelParameters.getFixedIndex()) + Tools.getLineSeparator() + Tools.getLineSeparator());
		
		double[][] stringCoordinates = modelParameters.getStringCoords();
		result.append("String image coordinates" + Tools.getLineSeparator());
		for (int i=0; i<stringCoordinates.length; i++) {
			result.append(String.format("%d ", i));
			for (int j=0; j<modelParameters.getMStringDimension(); j++) {
				result.append(String.format(" %.20f", stringCoordinates[i][j]));
			}
			result.append(Tools.getLineSeparator());
		}
		result.append(Tools.getLineSeparator());
		
		result.append("Scaling factors" + Tools.getLineSeparator());
		double[] scalingFactors = modelParameters.getScalingFactors();
		for (int i=0; i<scalingFactors.length;i++) {
			result.append(String.format("%d %.20f", i, scalingFactors[i]) + Tools.getLineSeparator());
		}
		result.append(Tools.getLineSeparator());
	
		result.append("Cubic spline mapping f(alpha)" + Tools.getLineSeparator());
		result.append(modelParameters.getFAlphaSpline().toString());
		result.append(Tools.getLineSeparator() + Tools.getLineSeparator());
		
		result.append(String.format("Final optimization score in model creation = %f", modelParameters.getFinalScore()) + Tools.getLineSeparator());
		
		return result.toString();
	}
	
	/**
	 * returns an array containing all names of all attributes used for training
	 */
	public String[] getAttributeNames() {
		return attributeNames;
	}
	
	/**
	 *  returns an array containing only the names of those attributes
	 * 	that have been selected to be included into the model
	 */
	public String[] getSelectedAttributeNames() {
		String[] attributeNames = new String[modelParameters.getMStringDimension()];
		int index = 0;
		for (int i = 0; i < selectedAttributes.length; i++) {
			if (selectedAttributes[i]) {
				attributeNames[index] = this.attributeConstructions[i];
				index++;
			}
		}
		return attributeNames;
	}
	
	public boolean[] getSelectedAttributes() {
		return selectedAttributes;
	}
	
	public String getFixedAttribute() {
		return fixedAttribute;
	}
	
}
