package cl;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.lang.Math;


/**
 * Performs the actual evaluation stuff for the string coordinate.
 * @hide
 */

public final class StringEvaluator {
	
	private int mStringPoints;
	private int mStringDimension;
	private double[][] stringCoords;
	private CubicSpline fAlphaSpline;
	private double[] scalingFactors;
	private double oneOverNMinusOne;

	// constructor with unpacked parameters, as long list
    public StringEvaluator(int mStringPoints, int mStringDimension, double[][] stringCoords, CubicSpline fAlphaSpline, double[] scalingFactors) {
    	this.mStringPoints = mStringPoints;
    	this.mStringDimension = mStringDimension;
    	this.stringCoords = stringCoords;
    	assert (mStringPoints + 2) == stringCoords.length;
    	this.fAlphaSpline = fAlphaSpline;
    	this.scalingFactors = scalingFactors;
    	this.oneOverNMinusOne = 1.0d / (mStringPoints-1);
    }
    
	// constructor with packed parameters
    public StringEvaluator(StringCoordinateModelParameters modelParameters) {
    	this.mStringPoints = modelParameters.getMStringPoints();
    	this.mStringDimension = modelParameters.getMStringDimension();
    	this.stringCoords = modelParameters.getStringCoords();
    	assert (mStringPoints + 2) == stringCoords.length;
    	this.fAlphaSpline = modelParameters.getFAlphaSpline();
    	this.scalingFactors = modelParameters.getScalingFactors();
    	this.oneOverNMinusOne = 1.0d / (mStringPoints-1);
    }
    
    public double getStringPrediction(double[] point) {
    	double alpha = getStringProgress(point);
    	if (alpha > 1.0d) alpha = 1.0d;
    	if (alpha < 0.0d) alpha = 0.0d;
    	double prediction = fAlphaSpline.interpolate(alpha);
    	return prediction;
    }
    
    /**
     * get the string as it is in the optimizer
     * @return a copy of the current string coordinates
     */
    public double[][] getStringCoords() {
    	double[][] newStringCoords = new double[mStringPoints+2][mStringDimension];
    	for (int i=0;i<mStringPoints+2;i++) {
    		for (int j=0;j<mStringDimension;j++) {
    			newStringCoords[i][j] = stringCoords[i][j];
    		}
    	}
    	return newStringCoords;
    }
    
    /**
     * get the string as it is in the optimizer
     * @return the _pointer_ to the current string coordinates
     */
    double[][] getStringCoordsPointer() {
    	return stringCoords;
    }
    
    public void setStringCoords(double[][] stringCoords) {
    	this.stringCoords = stringCoords;
    }
    
    /**
     * get the current state of f(alpha)
     * @return a copy of the current cubic spline
     */
    public CubicSpline getFAlphaSpline() {
    	double[] mX = fAlphaSpline.getX();
    	double[] mY = fAlphaSpline.getY();
    	CubicSpline newFAlphaSpline = CubicSpline.createMonotoneCubicSpline(mX, mY);
    	return newFAlphaSpline;
    }
    
    /**
     * get the current state of f(alpha)
     * @return the _pointer_ to the current cubic spline
     */
    CubicSpline getFAlphaSplinePointer() {
    	return fAlphaSpline;
    }
    
    public void setFAlphaSpline(CubicSpline fAlphaSpline) {
    	this.fAlphaSpline = fAlphaSpline;
    }
    
    /**
     * get scaling factors
     * @return a copy of the scaling factors for each dimension
     */
    public double[] getScalingFactors() {
    	double[] newScalingFactors = new double[mStringDimension];
    	for (int i=0;i<mStringDimension;i++) {
    		newScalingFactors[i] = scalingFactors[i];
    	}
    	return newScalingFactors;
    }
    
    /**
     * get scaling factors
     * @return a copy of the scaling factors for each dimension
     */
    double[] getScalingFactorsPointer() {
    	return scalingFactors;
    }
    
    public void setScalingFactor(double[] scalingFactors) {
    	this.scalingFactors = scalingFactors;
    }
         
    public double getStringProgress(double[] point) {
    
    	// start by finding closest string image
    	int[] closestIndizes = getClosestImage(point);
    	
    	// System.out.format("closestIndizes = (%d, %d, %d)\n", closestIndizes[0], closestIndizes[1], closestIndizes[2]);
    	
    	int sign;
    	
    	if (closestIndizes[0] < closestIndizes[1]) {
    		sign = -1;
    	}
    	else {
    		sign = 1;
    	}
    	
    	// System.out.format("sign = %d\n", sign);
    	
    	// now the vectors, notation as in J. Rogal et al., J. Chem. Phys. 133, 174109 (2010)
    	double[] v1 = new double[mStringDimension];
    	double[] v2 = new double[mStringDimension];
    	double[] v3 = new double[mStringDimension];
    	
    	for (int i=0;i<mStringDimension;i++) {
    		v1[i] = stringCoords[closestIndizes[0]][i] - point[i];
    		v2[i] = point[i] - stringCoords[closestIndizes[1]][i];
    		v3[i] = stringCoords[closestIndizes[2]][i] - stringCoords[closestIndizes[0]][i];
    	}
    	
    	double v11 = dotProduct(v1, v1);
    	double v13 = dotProduct(v1, v3);
    	double v22 = dotProduct(v2, v2);
    	double v33 = dotProduct(v3, v3);
    	
    	// System.out.format("v11 = %f\n", v11);
    	// System.out.format("v13 = %f\n", v13);
    	// System.out.format("v22 = %f\n", v22);
    	// System.out.format("v33 = %f\n", v33);
    	
    	double delta = 0.5d*(Math.sqrt(v13*v13 - v33*(v11 - v22)) - v13) / v33 - 0.5d;
    	// System.out.format("delta = %f\n", delta);
    	double alpha = oneOverNMinusOne * (closestIndizes[0] - 1 + sign*delta);
    	
    	return alpha;
    
    }
        
    int[] getClosestImage(double[] point) {
    	double dr2min = 1.0E100;
    	int imin = 0;
    	double[] dr2 = new double[mStringPoints+2];
    	for (int i=0;i<mStringPoints+2;i++) {
	    	for (int j=0;j<mStringDimension;j++) {
	    		dr2[i] += (point[j]-stringCoords[i][j]) * (point[j]-stringCoords[i][j]);
	    	}
	    	if (dr2[i] < dr2min) {
	    		dr2min = dr2[i];
	    		imin = i;
	    	}
    	}
    	int[] closestIndizes = new int[3];
    	if (imin==0) {
    		closestIndizes[0] = 1;
    		if (dr2[1] < dr2[2]) {
    			closestIndizes[1] = 0;
    			closestIndizes[2] = 2;
    		}
    		else {
    			closestIndizes[1] = 2;
    			closestIndizes[2] = 0;    			
    		}
    	}
    	else if (imin==(mStringPoints+1)) {
    		closestIndizes[0] = mStringPoints;
    		if (dr2[mStringPoints] < dr2[mStringPoints - 1]) { 
    			closestIndizes[1] = mStringPoints + 1;
    			closestIndizes[2] = mStringPoints - 1;
    		}
    		else {
    			closestIndizes[1] = mStringPoints - 1;
    			closestIndizes[2] = mStringPoints + 1;    			
    		}
    	}
    	else {
    		if (dr2[imin+1] < dr2[imin-1]) {
        		closestIndizes[0] = imin;
        		closestIndizes[1] = imin + 1;
        		closestIndizes[2] = imin - 1;
    		}
    		else {
        		closestIndizes[0] = imin;
        		closestIndizes[1] = imin - 1;
        		closestIndizes[2] = imin + 1;
    		}
    	}
    	return closestIndizes;
    }
    
    private double dotProduct(double[] a, double[] b) {
    	assert a.length == b.length;
    	double product = 0.0d;
    	for (int i=0;i<a.length;i++) {
    		product += a[i]*b[i];
    	}
    	return product;
    }
 
}
