package cl;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.lang.Math;
import java.util.Random;

import com.rapidminer.example.ExampleSet;
import com.rapidminer.tools.LoggingHandler;

/**
 * Performs the actual optimization for the string coordinate.
 * @hide
 */

public final class StringOptimizer {
	
	private int mStringPoints;
	private int mStringDimension;
	private final int fixedIndex;
	private double[][] coordinates;
	private double[] values;
	private int[] nA;
	private int[] nB;
	private double oneOverNMinusOne;
	private StringEvaluator stringEvaluator;
	private boolean useNANB;
	
	private Random random = new Random();
	
	private static final double STRING_STEP = 0.1d;
	private static final double REPARA_PRECISION = 1E-4d;
	private static final double ALPHA_STEP = 0.1d;
	private static final double SCALING_STEP = 0.1d;

    public StringOptimizer(int mStringPoints, int mStringDimension, double[][] stringCoords, CubicSpline fAlphaSpline, double[] scalingFactors, int fixedIndex, double[][] coordinates, double[] values) {
    	this.mStringPoints = mStringPoints;
    	this.mStringDimension = mStringDimension;
    	assert (mStringPoints + 2) == stringCoords.length;
    	this.fixedIndex = fixedIndex;
    	assert coordinates.length == values.length;
    	this.coordinates = coordinates;
    	this.nA = null;
    	this.nB = null;
    	this.values = values;
    	this.oneOverNMinusOne = 1.0d / (mStringPoints-1);
    	this.stringEvaluator = new StringEvaluator(mStringPoints, mStringDimension, stringCoords, fAlphaSpline, scalingFactors);
    	this.useNANB = false;
    }
    
    public StringOptimizer(int mStringPoints, int mStringDimension, double[][] stringCoords, CubicSpline fAlphaSpline, double[] scalingFactors, int fixedIndex, double[][] coordinates, int[] nA, int[] nB) {
    	this.mStringPoints = mStringPoints;
    	this.mStringDimension = mStringDimension;
    	assert (mStringPoints + 2) == stringCoords.length;
    	this.fixedIndex = fixedIndex;
    	assert coordinates.length == nA.length;
    	assert nA.length == nB.length;
    	this.coordinates = coordinates;
    	this.values = null;
    	this.nA = nA;
    	this.nB = nB;
    	this.oneOverNMinusOne = 1.0d / (mStringPoints-1);
    	this.stringEvaluator = new StringEvaluator(mStringPoints, mStringDimension, stringCoords, fAlphaSpline, scalingFactors);
    	this.useNANB = true;
    }
    
    public StringCoordinateModelParameters getStringCoordinateModelParameters() {
    	double[][] stringCoordinates = stringEvaluator.getStringCoords();
    	CubicSpline fAlphaSpline = stringEvaluator.getFAlphaSpline();
    	double[] scalingFactors = stringEvaluator.getScalingFactors();
    	StringCoordinateModelParameters modelParameters = new StringCoordinateModelParameters(mStringPoints, mStringDimension, fixedIndex, stringCoordinates, fAlphaSpline, scalingFactors);
    	return modelParameters;
    }
    
    public double optimize(int nSteps) {
    	return optimize(nSteps, null);
    }
        
    public double optimize(int nSteps, LoggingHandler loggingHandler) {
    	double newValue;
    	double oldValue = objectiveFunction();
    	
    	System.out.format("initial objective function value = %f\n", oldValue);
    	
    	double[][] stringCoords = stringEvaluator.getStringCoordsPointer();
    	
		for (int i=0;i<10;i++) {
			moveStringToMeanDistance(stringCoords, loggingHandler);
			fullyReparametrizeString(stringCoords);
		}
     	
		oldValue = objectiveFunction();
    	System.out.format("initial objective function value after mean distance move = %f\n", oldValue);
    	for (int step=0;step<nSteps;step++) {
    		
    		int randomInteger = random.nextInt(200);
    		
    		// string move
    		if (randomInteger < 10) {
        		double[][] tempStringCoords = stringEvaluator.getStringCoords();
            	stringCoords = stringEvaluator.getStringCoordsPointer();
        		randomStringMove(stringCoords);
        		fullyReparametrizeString(stringCoords);
        		newValue = objectiveFunction();
        		if (newValue < oldValue) {
        			oldValue = newValue;
        		}
        		else {
        			stringEvaluator.setStringCoords(tempStringCoords);
        		}
    		}
    		
    		// scaling move
    		else if (randomInteger < 20) {
	    		double[][] tempStringCoords = stringEvaluator.getStringCoords();
	        	double[][] tempCoordinates = new double[coordinates.length][mStringDimension];
	    		for (int i=0;i<coordinates.length;i++) {
	    			for (int j=0;j<mStringDimension;j++) {
	    				tempCoordinates[i][j] = coordinates[i][j];
	    			}
	    		}
	    		double[] scalingFactors = stringEvaluator.getScalingFactorsPointer();
	    		ScalingFactor oldScalingFactor = scalingMove(scalingFactors);
	    		newValue = objectiveFunction();
	      		if (newValue < oldValue) {
	    			oldValue = newValue;
	    		}
	    		else {
	    			stringEvaluator.setStringCoords(tempStringCoords);
	    			coordinates = tempCoordinates;
	    			scalingFactors[oldScalingFactor.changedIndex] = oldScalingFactor.scalingFactor;
	    		}
    		}
    		
    		// f_alpha move
    		else {
    			CubicSpline fAlphaSpline = stringEvaluator.getFAlphaSplinePointer();
    			CubicSpline oldSpline = stringEvaluator.getFAlphaSpline();
    			boolean preAccept = splineMove(fAlphaSpline);
	    		if (preAccept) {
	    			newValue = objectiveFunction();
	        		if (newValue < oldValue) {
	        			oldValue = newValue;
	        		}
	        		else {
	        			stringEvaluator.setFAlphaSpline(oldSpline);
	        		}
	    		}
    		}
    		
    		if (step%1000 == 0) {
    			System.out.format("%05d %f\n", step, oldValue);
    		}

    	}
    	System.out.format("final objective function value = %f\n", oldValue);
    	return oldValue;
    }
    
    private double objectiveFunction() {
    	double result;
    	if (useNANB) {
    		result = getLogLikelihood();
    	}
    	else {
    		result = getSquaredDeviation();
    	}
    	return result;
    }
    
    /**
     * simple test function to compare string prediction with real value
     * @return total quadratic deviation of prediction to real value
     */
    private double getSquaredDeviation() {
    	double sumDeviation = 0.0d;
    	for (int i=0;i<coordinates.length;i++) {
    		double prediction = stringEvaluator.getStringPrediction(coordinates[i]);
    		double delta = prediction - values[i];
    		sumDeviation += delta*delta;
    	}
    	return sumDeviation;
    }
    
    /**
     * model reaction coordinate for committor calculations
     * @return the negative log likelihood of the model reaction coordinate
     */
    private double getLogLikelihood() {
    	double logLikelihood = 0.0d;
    	for (int i=0;i<coordinates.length;i++) {
    		double prediction = stringEvaluator.getStringPrediction(coordinates[i]);
    		double modelPB = 0.5d + 0.5d*Math.tanh(prediction);
    		double likelihoodTerm = nB[i]*Math.log(modelPB) + nA[i]*Math.log(1.0d - modelPB);
    		logLikelihood -= likelihoodTerm;
    	}
    	return logLikelihood;
    }
        
    /**
     * perform a random movement of a single string image
     * @return the index of the moved string image
     */
    private int randomStringMove(double[][] stringCoords) {
    	// get inner string point = index from 1 to mStringPoints
    	// nextInt(n) returns random integer in [0, n)
    	int randomIndex = random.nextInt(mStringPoints) + 1;
    	if ((randomIndex > 1) && (randomIndex < mStringPoints)) {
    		// "normal" point: move in every direction
    		for (int i=0;i<mStringDimension;i++) {
    			double randomDelta = (random.nextDouble() - 0.5d)*STRING_STEP;
    			stringCoords[randomIndex][i] += randomDelta;
    		}
    	}
    	else {
    		// "fixed" point: do not move in fixed dimension
    		for (int i=0;i<mStringDimension;i++) {
    			if (i != fixedIndex) {
    				double randomDelta = (random.nextDouble() - 0.5d)*STRING_STEP;
    				stringCoords[randomIndex][i] += randomDelta;
    			}
    		}
    	}
    	return randomIndex;
    }
    
    /**
     * modify the cubic spline, the mapping from alpha (string process) to the actual coordinate
     * @return the original spline for the case the move is rejected later
     */
    private boolean splineMove(CubicSpline fAlphaSpline) {
    	double[] fAlphaX = fAlphaSpline.getX();
    	double[] fAlphaValues = fAlphaSpline.getY();
    	double[] fAlphaValuesNew = new double[mStringPoints];
    	for (int i=0;i<mStringPoints;i++) {
    		fAlphaValuesNew[i] = fAlphaValues[i];
    	}
    	// get a random index in [0, mStringPoints)
    	int randomIndex = random.nextInt(mStringPoints);
    	double randomDelta = (random.nextDouble() - 0.5d)*ALPHA_STEP;
    	double trialValue = fAlphaValues[randomIndex] + randomDelta;
    	boolean accept = false;
    	// now check if the proposed mapping is still monotonic
    	if (randomIndex==0) {
    		if (trialValue < fAlphaValues[1]) accept = true;
    	}
    	else if (randomIndex==mStringPoints-1) {
    		if (trialValue > fAlphaValues[mStringPoints-2]) accept = true;
    	}
    	else {
    		if ((trialValue > fAlphaValues[randomIndex-1]) && (trialValue < fAlphaValues[randomIndex+1])) accept = true;
    	}
    	if (accept) {
    		fAlphaValuesNew[randomIndex] = trialValue;
    		fAlphaSpline = CubicSpline.createMonotoneCubicSpline(fAlphaX, fAlphaValuesNew);
    		stringEvaluator.setFAlphaSpline(fAlphaSpline);
    	}
    	return accept;
    }
    
    /**
     * scale the system, i. e. scale a) all data points and b) the string itself
     * @return the original scaling factor in case the move has to be rejected
     */
    private ScalingFactor scalingMove(double[] scalingFactors) {
    	// get random index (but not the fixed, so that has to be skipped without destroying uniformity
    	int randomIndex = random.nextInt(mStringDimension-1);
    	if (randomIndex >= fixedIndex) randomIndex++;	// circumvent the fixed one
    	ScalingFactor oldScalingFactor = new ScalingFactor(randomIndex, scalingFactors[randomIndex]);
    	double randomDelta = (random.nextDouble() - 0.5d)*SCALING_STEP;
    	double relativeScaling = 1.0d + randomDelta;
    	scalingFactors[randomIndex] *= relativeScaling;
    	// scale the string first
    	double[][] stringCoords = stringEvaluator.getStringCoordsPointer();
    	for (int i=0;i<mStringPoints+2;i++) {
    		stringCoords[i][randomIndex] *= relativeScaling;
    	}
    	// now comes the bigger fish: the data points
    	for (int i=0;i<coordinates.length;i++) {
    		coordinates[i][randomIndex] *= relativeScaling;
    	}
    	return oldScalingFactor;
    }
    
    /**
     * re-parametrize string such that is equidistant again: place new images
     * on connection lines between old images, see also description in
     * L. Maragliano and E. Vanden-Eijnden, Phys. Chem. Lett. 446, 182-190 (2007)
     * @return the mean distance between adjacent string images
     */
    private double reparametrizeString(double[][] stringCoords) {
    	double[][] newStringCoords = new double[mStringPoints+2][mStringDimension];
		for (int i=0;i<mStringDimension;i++) {
			newStringCoords[1][i] = stringCoords[1][i];
			newStringCoords[mStringPoints][i] = stringCoords[mStringPoints][i];
		}
		double[] distances = new double[mStringPoints+2];
		double[] sumDistances = new double[mStringPoints+2];
		double[] distVector = new double[mStringDimension];
		for (int i=2;i<mStringPoints+1;i++) {
			for (int j=0;j<mStringDimension;j++) {
				distVector[j] =  stringCoords[i][j] - stringCoords[i-1][j];
			}
			double distSquared = dotProduct(distVector, distVector);
			distances[i] = Math.sqrt(distSquared);
			sumDistances[i] = sumDistances[i-1] + distances[i]; 
		}
		double meanDistance = oneOverNMinusOne * sumDistances[mStringPoints];
		// now loop over new string
		for (int i=1;i<mStringPoints+1;i++) {
			double meanSum = (i-1)*meanDistance;
			for (int j=2;j<mStringPoints+1;j++) {
				if (meanSum < sumDistances[j]) {
					for (int k=0;k<mStringDimension;k++) {
						newStringCoords[i][k] = stringCoords[j-1][k]  + (meanSum - sumDistances[j-1])/distances[j]*(stringCoords[j][k] - stringCoords[j-1][k]);	
					}
					break;
				}
				else {
					for (int k=0;k<mStringDimension;k++) {
						newStringCoords[i][k] = stringCoords[j][k];	
					}
				}
			}
		}
		adjustStringEnds(newStringCoords);
		stringEvaluator.setStringCoords(newStringCoords);
    	return meanDistance;
    }
    
    /**
     * to create initial string, move each string image in direction of its mean distance vector
     */
    private void moveStringToMeanDistance(double[][] stringCoords) {
    	moveStringToMeanDistance(stringCoords, null);
    }
    
    private void moveStringToMeanDistance(double[][] stringCoords, LoggingHandler loggingHandler) {
    	double[][] meanDistanceVectors = new double[mStringPoints+2][mStringDimension];
    	double[] weights = new double[mStringPoints+2];
    	for (int i=0;i<coordinates.length;i++) {
    		int[] closestIndizes = stringEvaluator.getClosestImage(coordinates[i]);
    		int closestIndex = closestIndizes[0];
    		double[] distanceVector = new double[mStringDimension];
    		for (int j=0;j<mStringDimension;j++) {
    			distanceVector[j] = coordinates[i][j] - stringCoords[closestIndex][j];
    			if (useNANB) {
        			meanDistanceVectors[closestIndex][j] += nA[i]*distanceVector[j];
        			meanDistanceVectors[closestIndex][j] += nB[i]*distanceVector[j];
    			}
    			else {
        			meanDistanceVectors[closestIndex][j] += distanceVector[j];
    			}
    		}
    		if (useNANB) {
    			weights[closestIndex] += (double) nA[i];
    			weights[closestIndex] += (double) nB[i];
    		}
    		else {
				weights[closestIndex] += 1.0d;
			}
    	}
    	for (int i=0;i<mStringPoints+2;i++) {
    		for (int j=0;j<mStringDimension;j++) {
    			if (weights[i] > 0.0d) {
    				meanDistanceVectors[i][j] = meanDistanceVectors[i][j] / weights[i];
    			}
    			else {
					if ( (loggingHandler != null) && (i != 0) && (i != (mStringPoints+1)) ) {
						String message = String.format("String image %d has zero closest points. Normalize your data points to the interval [0, 1].", i);
						loggingHandler.logWarning(message);
					}
				}
    		}
    	}
    	meanDistanceVectors[1][fixedIndex] = 0.0d;
    	meanDistanceVectors[mStringPoints][fixedIndex] = 0.0d;
    	for (int i=1;i<mStringPoints+1;i++) {
    		for (int j=0;j<mStringDimension;j++) {
    			stringCoords[i][j] += meanDistanceVectors[i][j];
    		}
    	}    	
    }
    
    private int fullyReparametrizeString(double[][] stringCoords) {
    	for (int i=0;i<30;i++) {
    		double meanDistance = reparametrizeString(stringCoords);
    		stringCoords = stringEvaluator.getStringCoordsPointer();
    		// calculate new mean distance
    		double[] distVector = new double[mStringDimension];
    		double newDistance = 0.0d;
    		for (int j=2;j<mStringPoints+1;j++) {
    			for (int k=0;k<mStringDimension;k++) {
    				distVector[k] =  stringCoords[j][k] - stringCoords[j-1][k];
    			}
    			double distSquared = dotProduct(distVector, distVector);
    			newDistance += Math.sqrt(distSquared);
    		}
    		newDistance -= (mStringPoints-1)*meanDistance;
    		if (Math.abs(newDistance) < REPARA_PRECISION) {
    			return 0;
    		}
    	}
    	System.out.format("Warning: string reparametrization failed.\n");
    	return 1;
    }
    
    // helper points at the end: just make straight line extension
    private void adjustStringEnds(double[][] stringCoords) {
    	for (int i=0;i<mStringDimension;i++) {
    		stringCoords[0][i] = 2.0d*stringCoords[1][i] - stringCoords[2][i];
    		stringCoords[mStringPoints+1][i] = 2.0d*stringCoords[mStringPoints][i] - stringCoords[mStringPoints-1][i];
    	}
    }
    
    private double dotProduct(double[] a, double[] b) {
    	assert a.length == b.length;
    	double product = 0.0d;
    	for (int i=0;i<a.length;i++) {
    		product += a[i]*b[i];
    	}
    	return product;
    }
 
}
