/*
 *  RapidMiner
 *
 *  Copyright (C) 2001-2015 by RapidMiner GmbH and the contributors
 *
 *  Complete list of developers available at our web site:
 *
 *       http://rapidminer.com
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see http://www.gnu.org/licenses/.
 */
package cl.stringcoordinate;

import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import Jama.Matrix;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.AttributeWeights;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.Statistics;
import com.rapidminer.example.set.SimpleExampleSet;
import com.rapidminer.example.table.AttributeFactory;
import com.rapidminer.example.table.DataRow;
import com.rapidminer.example.table.DoubleArrayDataRow;
import com.rapidminer.example.table.MemoryExampleTable;
import com.rapidminer.operator.Model;
import com.rapidminer.operator.OperatorCapability;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.annotation.ResourceConsumptionEstimator;
import com.rapidminer.operator.learner.AbstractLearner;
import com.rapidminer.operator.learner.PredictionModel;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.preprocessing.join.ExampleSetMerge;
import com.rapidminer.operator.preprocessing.join.ExampleSetSuperset;
import com.rapidminer.operator.visualization.dependencies.NumericalMatrix;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeAttribute;
import com.rapidminer.parameter.ParameterTypeBoolean;
import com.rapidminer.parameter.ParameterTypeCategory;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.parameter.UndefinedParameterError;
import com.rapidminer.tools.LoggingHandler;
import com.rapidminer.tools.Ontology;
import com.rapidminer.tools.OperatorResourceConsumptionHandler;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.Tools;
import com.rapidminer.tools.math.MathFunctions;

import cl.CubicSpline;
import cl.StringCoordinateModel;
import cl.StringCoordinateModelParameters;
import cl.StringEvaluator;
import cl.StringOptimizer;

/**
 * <p>
 * This operator applies the string coordinate. Original paper: W. Lechner, ...
 * Select a column for
 * </p>
 * 
 * @author Christian Leitold
 */
public class StringCoordinate extends AbstractLearner {

    /**
     * The parameter name for &quot;The feature selection method used during
     * regression.&quot;
     */
    public static final String PARAMETER_OPTIMIZATION_METHOD = "optimization_method";

    public static final Map<String, Class<? extends StringOptimizer>> OPTIMIZATION_METHODS = new LinkedHashMap<String, Class<? extends StringOptimizer>>();

    static {
    	OPTIMIZATION_METHODS.put("simple", StringOptimizer.class);
    	OPTIMIZATION_METHODS.put("simulated annealing (not implemented yet)", StringOptimizer.class);
    }

    /** Standard optimization method: Wolfgang Lechner */
    public static final int SIMPLE = 0;

    /** Simulated annealing as optimization method */
    public static final int SIMULATED_ANNEALING = 1;
    
    /**
     * The parameter name for &quot;The attribute to keep fixed.&quot;
     */
    public static final String PARAMETER_FIXED_ATTRIBUTE = "fixed_attribute";
    
    /**
     * How many string images should be used? (default: 5)
     */
    public static final String PARAMETER_NUMBER_OF_STRING_IMAGES = "number of string images";
    
    /**
     * Perform committor calculation / string reaction coordinate?;
     */
    public static final String PARAMETER_USE_COMMITTORDATA = "use_committor_data";
    
    /**
     * The actual attributes that transport the NA / NB data.;
     */
    public static final String PARAMETER_ATTRIBUTE_NA = "attribute_for_NA";
    public static final String PARAMETER_ATTRIBUTE_NB = "attribute_for_NB";
    
    /**
     * Switch if string should be reversed (mainly for committor calculation)
     */
    public static final String PARAMETER_REVERSE_STRING = "reverse_string";

    /**
     * Switch if string should be stretched to fit unnormalized data
     */
    public static final String PARAMETER_STRETCH_STRING = "stretch_string_to_data_range";
    
    /**
     * Number of iterations in optimization procedure
     */
    public static final String PARAMETER_NUMBER_OF_ITERATIONS = "number_of_iterations";
    
    /**
     * Flag to "predict" reaction coordinate value instead of committor (differs by application of tanh)
     */
    // public static final String PARAMETER_CALCULATE_REACTION_COORDINATE = "calculate_reaction_coordinate";
    
    private OutputPort weightOutput = getOutputPorts().createPort("weights");
    
    private OutputPort stringCoordinateOutput = getOutputPorts().createPort("stringCoordinateVisualization");
    
    private OutputPort finalScoreOutput = getOutputPorts().createPort("finalScore");

    public StringCoordinate(OperatorDescription description) {
        super(description);

        getTransformer().addGenerationRule(weightOutput, AttributeWeights.class);
        
        // now for the string stuff
        getTransformer().addGenerationRule(stringCoordinateOutput, ExampleSet.class);
    }

    @Override
    public Model learn(ExampleSet exampleSet) throws OperatorException {
        // initializing data and parameter values.
        Attribute label = exampleSet.getAttributes().getLabel();
        Attribute workingLabel = label;
        boolean cleanUpLabel = false;
                
        String fixedAttribute = getParameterAsString(PARAMETER_FIXED_ATTRIBUTE);
        
        int numberOfStringImages = getParameterAsInt(PARAMETER_NUMBER_OF_STRING_IMAGES);
        
        boolean useCommittorData = getParameterAsBoolean(PARAMETER_USE_COMMITTORDATA);
        
        String attributeNA = getParameterAsString(PARAMETER_ATTRIBUTE_NA);
        String attributeNB = getParameterAsString(PARAMETER_ATTRIBUTE_NB);
        
        boolean reverseString = getParameterAsBoolean(PARAMETER_REVERSE_STRING);
        
        boolean stretchString = getParameterAsBoolean(PARAMETER_STRETCH_STRING);
        
        int numberOfIterations = getParameterAsInt(PARAMETER_NUMBER_OF_ITERATIONS);
        
        // boolean calculateReactionCoordinate = getParameterAsBoolean(PARAMETER_CALCULATE_REACTION_COORDINATE);
        
        // prepare for classification by translating into 0-1 coding.
        if (label.isNominal()) {
            if (label.getMapping().size() == 2) {
                
                int firstIndex = label.getMapping().getNegativeIndex();

                // here, we create a new attribute workingLabel which is used for the regression getLog()
                workingLabel = AttributeFactory.createAttribute("regression_label", Ontology.REAL);
                exampleSet.getExampleTable().addAttribute(workingLabel);

                // loop over all examples in the example set and assign workingLabel = 0.0 or 1.0
                for (Example example : exampleSet) {
                    double index = example.getValue(label);
                    if (index == firstIndex) {
                        example.setValue(workingLabel, 0.0d);
                    } else {
                        example.setValue(workingLabel, 1.0d);
                    }
                }

                exampleSet.getAttributes().setLabel(workingLabel);
                cleanUpLabel = true;
                
            }
        }
        
        // search all attributes and keep numerical (all numerical are USED, except NA and NB if defined)
        int numberOfAttributes = exampleSet.getAttributes().size();
        boolean[] isUsedAttribute = new boolean[numberOfAttributes];
        int counter = 0;
        String[] attributeNames = new String[numberOfAttributes];
        for (Attribute attribute : exampleSet.getAttributes()) {
        	if (useCommittorData) {
        		String attributeName = attribute.getName();
        		boolean notNANB = !(attributeName.equals(attributeNA) || attributeName.equals(attributeNB)); 
        		isUsedAttribute[counter] = attribute.isNumerical() && notNANB;
        	}
        	else {
        		isUsedAttribute[counter] = attribute.isNumerical();
        	}
        	attributeNames[counter] = attribute.getName();
        	counter++;  
        }

        // compute and store statistics and turn off attributes with zero
        // standard deviation
        exampleSet.recalculateAllAttributeStatistics();
        double[] means = new double[numberOfAttributes];
        double[] standardDeviations = new double[numberOfAttributes];
        counter = 0;
        Attribute[] allAttributes = new Attribute[numberOfAttributes];
        for (Attribute attribute : exampleSet.getAttributes()) {
            allAttributes[counter] = attribute;
            if (isUsedAttribute[counter]) {
                means[counter] = exampleSet.getStatistics(attribute, Statistics.AVERAGE_WEIGHTED);
                standardDeviations[counter] = Math.sqrt(exampleSet.getStatistics(attribute, Statistics.VARIANCE_WEIGHTED));
                if (standardDeviations[counter] == 0) {
                    isUsedAttribute[counter] = false;
                    String message = String.format("String Coordinate: attribute %s is excluded (zero standard deviation).", attributeNames[counter]);
                    exampleSet.getLog().log(message + Tools.getLineSeparator());
                }
            }
            counter++;
        }

        double labelMean = exampleSet.getStatistics(workingLabel, Statistics.AVERAGE_WEIGHTED);

        // determine the number of used attributes + 1 (for label value)
        int numberOfUsedAttributes = 1;
        for (int i = 0; i < isUsedAttribute.length; i++) {
            if (isUsedAttribute[i]) {
                numberOfUsedAttributes++;
            }
        }

        int M = numberOfStringImages;
        int m = numberOfUsedAttributes - 1;
        
        StringCoordinateModelParameters modelParameters = null;
        if (useCommittorData) {
        	modelParameters = performRegression(exampleSet, isUsedAttribute, means, labelMean, M, m, fixedAttribute, reverseString, stretchString, numberOfIterations, useCommittorData, attributeNA, attributeNB);
        }
        else {
        	modelParameters  = performRegression(exampleSet, isUsedAttribute, means, labelMean, M, m, fixedAttribute, reverseString, stretchString, numberOfIterations);	
        }
        
        // model is calculated, now get coordinates of string images for visualization
        List<Attribute> stringImageAttributes = new LinkedList<Attribute>();
        Attribute newLabel = (Attribute) label.clone();
        stringImageAttributes.add(newLabel);
        counter = 0;
        for (Attribute attribute : exampleSet.getAttributes()) {
        	// System.out.format("... counter = %d, name = %s.\n", counter++, attribute.getName());
        	stringImageAttributes.add((Attribute) attribute.clone());
        }
        Attribute stringImage = AttributeFactory.createAttribute("string_image", Ontology.BINOMINAL);
        stringImageAttributes.add(stringImage);        
        MemoryExampleTable stringImageExampleTable = new MemoryExampleTable(stringImageAttributes);
        
        // add the coordinates of the string images to the new example set
        // unused variables get a value of -1
        // we scale the STRING (to minimize user confusion), even though in the optimizer the DATA are scaled
        double[][] stringCoordinates = modelParameters.getStringCoords();
        double[] scalingFactors = modelParameters.getScalingFactors();
        for (int i=0;i<M+2;i++) {
        	double[] data = new double[numberOfUsedAttributes+2];
        	counter = 0;
        	int dataCounter = 0;
        	for (int j=0;j<numberOfAttributes;j++) {
        		if (isUsedAttribute[j]) {
        			data[dataCounter+1] = stringCoordinates[i][counter] / scalingFactors[counter];
        			counter++;
        			dataCounter++;
        		}
        		else if (useCommittorData) {
        			if (attributeNames[j].equals(attributeNA) || attributeNames[j].equals(attributeNB)) {
        				data[dataCounter+1] = -1;
        				dataCounter++;
        			}
        		}
        		
        	}
        	DataRow dataRow = new DoubleArrayDataRow(data);
        	stringImageExampleTable.addDataRow(dataRow);
        }
        
        SimpleExampleSet stringImageExampleSet = new SimpleExampleSet(stringImageExampleTable);
        stringImageExampleSet.getAttributes().setLabel(newLabel);
        
        // set for all the string images the corresponding flag to "yes"
        for (Example example : stringImageExampleSet) {
        	example.setValue(stringImage, "yes");
        }
        
        // finally, get the "right" prediction for each point
        StringEvaluator stringEvaluator = new StringEvaluator(modelParameters);
        for (int i=1;i<stringCoordinates.length-1;i++) {
        	double prediction = stringEvaluator.getStringPrediction(stringCoordinates[i]);
        	if (useCommittorData) {
        		double pB = 0.5d + 0.5d*Math.tanh(prediction);
        		stringImageExampleSet.getExample(i).setValue(newLabel, pB);
        	}
        	else {
        		stringImageExampleSet.getExample(i).setValue(newLabel, prediction);
        	}
        }
        // the helper points have to be treated separately due to NaN occurrence
        // the chance that this occurs during the normal optimization has (almost) measure zero
        CubicSpline fAlphaSpline = modelParameters.getFAlphaSpline();
        if (useCommittorData) {
        	double prediction = fAlphaSpline.interpolate(0.0d);
        	double pB = 0.5d + 0.5d*Math.tanh(prediction);
        	stringImageExampleSet.getExample(0).setValue(newLabel, pB);
        	prediction = fAlphaSpline.interpolate(1.0d);
        	pB = 0.5d + 0.5d*Math.tanh(prediction);
        	stringImageExampleSet.getExample(stringCoordinates.length-1).setValue(newLabel, pB);
        }
        else {
        	double prediction = fAlphaSpline.interpolate(0.0d);
        	stringImageExampleSet.getExample(0).setValue(newLabel, prediction);    	
        	prediction = fAlphaSpline.interpolate(1.0d);
        	stringImageExampleSet.getExample(stringCoordinates.length-1).setValue(newLabel, prediction);
        }
        
        ExampleSet mergedExampleSet = null;
        try {
        	
        	ExampleSet exampleSet1 = (ExampleSet) stringImageExampleSet.clone();
    		ExampleSet exampleSet2 = (ExampleSet) exampleSet.clone();
        	
        	ExampleSetSuperset supersetOperator = OperatorService.createOperator(ExampleSetSuperset.class);
			supersetOperator.setParameter(ExampleSetSuperset.PARAMETER_INCLUDE_SPECIAL_ATTRIBUTES, "true");			
			supersetOperator.superset(exampleSet1, exampleSet2);
			
			ExampleSetMerge mergeOperator = OperatorService.createOperator(ExampleSetMerge.class);
			
			List<ExampleSet> exampleSetList = new LinkedList<ExampleSet>();
			exampleSetList.add(stringImageExampleSet);
			exampleSetList.add(exampleSet);
			
			mergedExampleSet = mergeOperator.merge(Arrays.asList(new ExampleSet[] { exampleSet1, exampleSet2} ));
						
		} catch (OperatorCreationException e1) {
			// Auto-generated catch block (just use that as-is)
			e1.printStackTrace();
		}
        

        // apply string coordinate method: preparations to get method
        int selectionMethodIndex = getParameterAsInt(PARAMETER_OPTIMIZATION_METHOD);
        String[] selectionMethodNames = OPTIMIZATION_METHODS.keySet().toArray(new String[OPTIMIZATION_METHODS.size()]);
        String selectedMethod = selectionMethodNames[selectionMethodIndex];
        Class<? extends StringOptimizer> optimizerClass = OPTIMIZATION_METHODS.get(selectedMethod);
        if (optimizerClass == null) {
            throw new UserError(this, 904, PARAMETER_OPTIMIZATION_METHOD, "unknown method");
        }
        
        /* StringOptimizer optimizer = null;
        try {
        	optimizer = optimizerClass.newInstance();
        } catch (InstantiationException e) {
            throw new UserError(this, 904, PARAMETER_OPTIMIZATION_METHOD, e.getMessage());
        } catch (IllegalAccessException e) {
            throw new UserError(this, 904, PARAMETER_OPTIMIZATION_METHOD, e.getMessage());
        } */

        // clean up eventually if was classification
        if (cleanUpLabel) {
            exampleSet.getAttributes().remove(workingLabel);
            exampleSet.getExampleTable().removeAttribute(workingLabel);
            exampleSet.getAttributes().setLabel(label);
        }

        // delivering weights
        if (weightOutput.isConnected()) {
            AttributeWeights weights = new AttributeWeights(exampleSet);
            for (int i = 0; i < attributeNames.length; i++) {
                if (isUsedAttribute[i]) {
                    weights.setWeight(attributeNames[i], (double) i + 1);
                } else {
                    weights.setWeight(attributeNames[i], 0);
                }
            }
            weightOutput.deliver(weights);
        }
        
        // delivering visualization for string coordinate
        if (stringCoordinateOutput.isConnected()) {
        	stringCoordinateOutput.deliver(mergedExampleSet);
        }
        
        if (finalScoreOutput.isConnected()) {
        	double finalScore = modelParameters.getFinalScore();
        	Matrix scoreMatrix = new Matrix(1, 1, finalScore);
        	String[] matrixName = {"finalScore"};
        	NumericalMatrix numericalMatrix = new NumericalMatrix("finalScore", matrixName, scoreMatrix, true);
        	finalScoreOutput.deliver(numericalMatrix);
        }

        return new StringCoordinateModel(exampleSet, isUsedAttribute, fixedAttribute, modelParameters, useCommittorData);
    }

    double getTolerance(ExampleSet exampleSet, boolean[] isUsedAttribute, int testAttributeIndex) throws UndefinedParameterError {
        List<Attribute> attributeList = new LinkedList<Attribute>();
        Attribute currentAttribute = null;
        int resultAIndex = 0;
        for (Attribute a : exampleSet.getAttributes()) {
            if (isUsedAttribute[resultAIndex]) {
                if (resultAIndex != testAttributeIndex) {
                    attributeList.add(a);
                } else {
                    currentAttribute = a;
                }
            }
            resultAIndex++;
        }

        Attribute[] usedAttributes = new Attribute[attributeList.size()];
        attributeList.toArray(usedAttributes);
        
        // this is dummy code
        double[] attributeValues = new double[exampleSet.size()];
        double[] predictedValues = new double[exampleSet.size()];
        int eIndex = 0;
        for (Example e : exampleSet) {
            attributeValues[eIndex] = e.getValue(currentAttribute);
            double prediction = 0.0d;
            for (Attribute a : usedAttributes) {
                prediction += 1.0d * e.getValue(a);
            }
            predictedValues[eIndex] = prediction;
            eIndex++;
        }

        double correlation = MathFunctions.correlation(attributeValues, predictedValues);
        double tolerance = 1.0d - correlation * correlation;
        return tolerance;
    }

    /** Calculates the squared error of a regression model on the training data. */
    double getSquaredError(ExampleSet exampleSet, boolean[] selectedAttributes, double[] coefficients) {
        double error = 0;
        Iterator<Example> i = exampleSet.iterator();
        while (i.hasNext()) {
            Example example = i.next();
            double prediction = regressionPrediction(example, selectedAttributes, coefficients);
            double diff = prediction - example.getLabel();
            error += diff * diff;
        }
        return error;
    }

    double getCorrelation(ExampleSet exampleSet, boolean[] selectedAttributes, double[] coefficients) {
        double[] labelValues = new double[exampleSet.size()];
        double[] predictions = new double[exampleSet.size()];
        int index = 0;
        for (Example e : exampleSet) {
            labelValues[index] = e.getLabel();
            predictions[index] = regressionPrediction(e, selectedAttributes, coefficients);
            index++;
        }
        return MathFunctions.correlation(labelValues, predictions);
    }

    /** Calculates the prediction for the given example. */
    private double regressionPrediction(Example example, boolean[] selectedAttributes, double[] coefficients) {
        double prediction = 0;
        int index = 0;
        int counter = 0;
        for (Attribute attribute : example.getAttributes()) {
            if (selectedAttributes[counter++]) {
                prediction += coefficients[index] * example.getValue(attribute);
                index++;
            }
        }

        return prediction;
    }
    
    /**
     * Calculate a string regression only from the selected attributes. The
     * method returns the calculated string parameters.
     * @param stretchString 
     */
    StringCoordinateModelParameters performRegression(ExampleSet exampleSet, boolean[] selectedAttributes, double[] means, double labelMean, int mStringPoints, int mStringDimension, String fixedAttribute, boolean reverseString, boolean stretchString, int numberOfIterations) throws UndefinedParameterError {
    	boolean useNANB = false;
    	return performRegression(exampleSet, selectedAttributes, means, labelMean, mStringPoints, mStringDimension, fixedAttribute, reverseString, stretchString, numberOfIterations, useNANB, "", "");
    }

    /**
     * Calculate a string regression only from the selected attributes. The
     * method returns the calculated string parameters.
     */
    @SuppressWarnings("null")
	StringCoordinateModelParameters performRegression(ExampleSet exampleSet, boolean[] selectedAttributes, double[] means, double labelMean, int mStringPoints, int mStringDimension, String fixedAttribute, boolean reverseString, boolean stretchString, int numberOfIterations, boolean useNANB, String attributeNA, String attributeNB) throws UndefinedParameterError {
        
    	// Lets pack the examples to plain double arrays
    	double[][] coordinates = null;
    	double[] values = null;
        // in case we have NA / NB information, they have to be packed to arrays as well
    	int[] nA = null;
    	int[] nB = null;
        if (mStringDimension > 1) {
            coordinates = new double[exampleSet.size()][mStringDimension];
            values = new double[exampleSet.size()];
        	if (useNANB) {
        		nA = new int[exampleSet.size()];
        		nB = new int[exampleSet.size()];
        	}
            int exampleIndex = 0;
            Iterator<Example> i = exampleSet.iterator();
            while (i.hasNext()) {
                Example example = i.next();
                int attributeIndex = 0;
                values[exampleIndex] = example.getLabel();
                int counter = 0;
                for (Attribute attribute : exampleSet.getAttributes()) {
                    if (selectedAttributes[counter]) { 
                        coordinates[exampleIndex][attributeIndex] = example.getValue(attribute);
                        attributeIndex++;
                    }
                    // nA and nB can never be null if this part is executed: controlled by useNANB
                    if (useNANB) {
                    	if (attribute.getName().equals(attributeNA)) {
                    		nA[exampleIndex] = (int) example.getValue(attribute);
                    	}
                    	if (attribute.getName().equals(attributeNB)) {
                    		nB[exampleIndex] = (int) example.getValue(attribute);
                    	}
                    }
                    counter++;
                }
                exampleIndex++;
            }
        }
            	
        // Now create the string and the mapping f(alpha)
        double[][] stringCoords = new double[mStringPoints+2][mStringDimension];
        double[] alpha = new double[mStringPoints];
        double[] fAlpha = new double[mStringPoints];
        
        // Let's start by creating the map f(alpha) as a CubicSpline
        
        double deltaAlpha = 1.0d / (mStringPoints-1);
        
        for (int i=0;i<mStringPoints;i++) {
        	alpha[i] = i*deltaAlpha;
        	fAlpha[i] = -5.0d + 10.0d*alpha[i];
        }
        
        CubicSpline fAlphaSpline = CubicSpline.createMonotoneCubicSpline(alpha, fAlpha);
        
        // The scaling factors for all the dimensions
        double[] scalingFactors = new double[mStringDimension];
        for (int i=0;i<mStringDimension;i++) {
        	scalingFactors[i] = 1.0d;
        }
        
        // Find out index of fixed coordinate / attribute: only consider USED attributes
        int counter = 0;
        int attributeCounter = 0;
        int fixedIndex = 0;
        for (Attribute attribute : exampleSet.getAttributes()) {
        	if (selectedAttributes[attributeCounter]) {
        		if (fixedAttribute.equals(attribute.getName())) {
            		fixedIndex = counter;
            		// System.out.format("StringCoordinate: fixed attribute is %s, fixed index is %d.\n", fixedAttribute, fixedIndex);
            	}
        		counter++;
        	}
            attributeCounter++;
        }
        
        /* 
         * Create the initial string: If the stretchString option is not set, the assumption is that the data range, especially for the fixed direction, is [0, 1]. In that case, the initial string will run from 0 to 1 in the fixed direction, and have a value of 0.5 everywhere else. If stretchedString is set (default), it will run from min to max in the fixed direction. The string direction can be reversed as well.
         */
        
        double stringImageMin, stringImageMax;
        
        if (stretchString) {
        	stringImageMin = coordinates[0][fixedIndex];
        	stringImageMax = coordinates[0][fixedIndex];
        	for (int i=0;i<coordinates.length;i++) {
        		if (coordinates[i][fixedIndex] < stringImageMin) {
        			stringImageMin = coordinates[i][fixedIndex];
        		}
        		if (coordinates[i][fixedIndex] > stringImageMax) {
        			stringImageMax = coordinates[i][fixedIndex];
        		}
        	}
        }
        else {
        	stringImageMin = 0.0d;
        	stringImageMax = 1.0d;        	
        }
        
        double deltaString = (stringImageMax - stringImageMin) / (mStringPoints-1);
        
        for (int i=0;i<mStringPoints+2;i++) {
        	for (int j=0;j<mStringDimension;j++) {
        		stringCoords[i][j] = 0.5d;
        	}
        	if (reverseString) {
        		stringCoords[i][fixedIndex] = stringImageMax - (i-1)*deltaString;
        	}
        	else {
        		stringCoords[i][fixedIndex] = stringImageMin + (i-1)*deltaString;
        	}
        }
        
        // Now we have to optimize the string: use extra class for that
        StringOptimizer optimizedString = null;
        if (useNANB) {
        	 optimizedString = new StringOptimizer(mStringPoints, mStringDimension, stringCoords, fAlphaSpline, scalingFactors, fixedIndex, coordinates, nA, nB);
        }
        else {
        	optimizedString = new StringOptimizer(mStringPoints, mStringDimension, stringCoords, fAlphaSpline, scalingFactors, fixedIndex, coordinates, values);
        }
        
        // optimizedString.testCode();
        LoggingHandler loggingHandler = exampleSet.getLog();
        String message = null;
        double finalValue = optimizedString.optimize(numberOfIterations, loggingHandler);
        if (useNANB) {
            message = String.format("String Coordinate: optimization finished with final log likelihood score of %f.", -finalValue);
        }
        else {
        	message = String.format("String Coordinate: optimization finished with final objection function value of %f.", finalValue);
        }
        loggingHandler.log(message);
        
        // get the model parameters as a whole from the optimizer
        StringCoordinateModelParameters parameters = optimizedString.getStringCoordinateModelParameters();
        
        parameters.setFinalScore(finalValue);
        
        return parameters;
    }

    @Override
    public Class<? extends PredictionModel> getModelClass() {
        return StringCoordinateModel.class;
    }

    @Override
    public boolean supportsCapability(OperatorCapability lc) {
        if (lc.equals(OperatorCapability.NUMERICAL_ATTRIBUTES))
            return true;
        if (lc.equals(OperatorCapability.NUMERICAL_LABEL))
            return true;
        if (lc.equals(OperatorCapability.BINOMINAL_LABEL))
            return true;
        if (lc == OperatorCapability.WEIGHTED_EXAMPLES)
            return false;
        return false;
    }

    @Override
    public List<ParameterType> getParameterTypes() {
        List<ParameterType> types = super.getParameterTypes();
        
        String[] availableOptimizationMethods = OPTIMIZATION_METHODS.keySet().toArray(new String[OPTIMIZATION_METHODS.size()]);

        types.add(new ParameterTypeCategory(PARAMETER_OPTIMIZATION_METHOD, "Indicates which method to use during the optimization of the string.", availableOptimizationMethods, SIMPLE));
        
        types.add(new ParameterTypeAttribute(PARAMETER_FIXED_ATTRIBUTE, "Indicates which attribute is used to fix the string.", getExampleSetInputPort()));
        
        types.add(new ParameterTypeInt(PARAMETER_NUMBER_OF_STRING_IMAGES, "Indicates the number of string images to use.", 2, 50, 5));
        
        types.add(new ParameterTypeBoolean(PARAMETER_USE_COMMITTORDATA, "Indicates whether committor data should be used, you have to specify attributes for NA and NB in that case.", false, false));
        
        types.add(new ParameterTypeAttribute(PARAMETER_ATTRIBUTE_NA, "Attribute name for NA data.", getExampleSetInputPort(), true, false));
        types.add(new ParameterTypeAttribute(PARAMETER_ATTRIBUTE_NB, "Attribute name for NB data.", getExampleSetInputPort(), true, false));
        
        types.add(new ParameterTypeBoolean(PARAMETER_REVERSE_STRING, "Indicates if direction of string should be reversed.", false, false));
        
        types.add(new ParameterTypeBoolean(PARAMETER_STRETCH_STRING, "Indicates if the extension of the string should be fitted to unnormalized data.", true, false));
        
        types.add(new ParameterTypeInt(PARAMETER_NUMBER_OF_ITERATIONS, "Indicates the number of iterations to use in the string optimization procedure.", 0, 2000000, 20000, false));
        
        // types.add(new ParameterTypeBoolean(PARAMETER_CALCULATE_REACTION_COORDINATE, "Indicates if reaction coordinate should be predicted instead of committor value. This is only evaluated if committor data NA / NB are used.", false, true));
        
        return types;
    }

    @Override
    public ResourceConsumptionEstimator getResourceConsumptionEstimator() {
        return OperatorResourceConsumptionHandler.getResourceConsumptionEstimator(getExampleSetInputPort(), StringCoordinate.class, null);
    }
}
